open Tags
open Syntax

let rec check_wf x seen = function
  | Eps | Elem _ -> ()
  | Seq (r1,r2) | Alt (r1,r2) -> check_wf x seen r1; check_wf x seen r2
  | Star r | Plus r -> check_wf x seen r
  | Name y when x = y -> failwith ("Non-linear: " ^ x)
  | Name y when List.mem y seen -> ()
  | Name y -> ignore (def y); check_wf x (y::seen) (def x)
  | e -> failwith ("Not a regexp: " ^ Syntax.to_string e)

let name_checked = Hashtbl.create 64

let rec check = function
  | Eps -> ()
  | Elem (t,c) -> check c
  | Seq (r1,r2) | Alt (r1,r2) -> check r1; check r2
  | Star r | Plus r -> check r
  | Name x when Hashtbl.mem name_checked x -> ()
  | Name x -> 
      let d = Syntax.def x in 
      check_wf x [] d; 
      Hashtbl.add name_checked x ();
      check d
  | e -> failwith ("Not a regexp: " ^ Syntax.to_string e)


let rec print pri ppf = function
  | Name s -> Format.fprintf ppf "%s" s
  | Seq (r1,r2) -> 
      if pri >= 2 then Format.fprintf ppf "(";
      Format.fprintf ppf "%a,%a" (print 1) r1 (print 1) r2;
      if pri >= 2 then Format.fprintf ppf ")"
  | Alt (r1,r2) -> 
      if pri >= 1 then Format.fprintf ppf "(";
      Format.fprintf ppf "%a|%a" (print 0) r1 (print 0) r2;
      if pri >= 1 then Format.fprintf ppf ")"
  | Star r -> 
      Format.fprintf ppf "%a*" (print 2) r
  | Plus r -> 
      Format.fprintf ppf "%a+" (print 2) r
  | Eps ->
      Format.fprintf ppf "()"
  | Elem (sset,sub) ->
      Format.fprintf ppf "%a[%a]" SymbolSet.print sset (print 0) sub
  | _ -> assert false

let print = print 0

let memo = Hashtbl.create 1024

let rec nillable = function
  | Eps | Star _ -> true
  | Seq (r1,r2) -> nillable r1 && nillable r2
  | Alt (r1,r2) -> nillable r1 || nillable r2
  | Plus r -> nillable r
  | Elem _ -> false
  | Name x -> nillable (def x)
  | _ -> assert false

let rec compile r =
  try Hashtbl.find memo r
  with Not_found ->
    let q = State.mk () in
    Hashtbl.add memo r q;
    State.link q (aux State.empty_seq r);
    q

and aux q = function
  | Eps -> q
  | Seq (r1,r2) -> aux (aux q r2) r1
  | Alt (r1,r2) -> State.alt (aux q r1) (aux q r2)
  | Star r ->
      let q' = State.mk () in
      State.link q' (State.alt (aux q' r) q);
      q'
  | Plus r -> (* aux (aux q (Star r)) r *)
      if nillable r then aux q (Star r)
      else
	let q' = State.mk () in
	let q'' = aux q' r in
	State.link q' (State.alt q'' q);
	q''
  | Elem (sset,r) -> State.cons false [ sset, compile r, q ]
  | Name x -> aux q (def x)
  | _ -> assert false

module D = Decompile.Make
  (Hashtbl.Make(State))
  (struct
     type t = SymbolSet.t * State.t
     let equal (sset1,q1) (sset2,q2) =
       SymbolSet.equal sset1 sset2 && State.equal q1 q2
     let compare (sset1,q1) (sset2,q2) =
       let c = SymbolSet.compare sset1 sset2 in
       if c != 0 then c else State.compare q1 q2
     let hash (sset,q) = SymbolSet.hash sset + 17 * State.hash q
   end)


module StateHash = Hashtbl.Make(State)

let state_names = StateHash.create 1024

let rec decompile q =
  if StateHash.mem state_names q then Name "LOOP"
  else 
    let () = StateHash.add state_names q () in
    let re = D.run State.auto q in
    let rec aux = function
      | Decompile.Epsilon -> Eps
      | Decompile.Seq (r1,r2) -> Seq (aux r1, aux r2)
      | Decompile.Alt (r1,r2) -> Alt (aux r1, aux r2)
      | Decompile.Star r -> Star (aux r)
      | Decompile.Plus r -> Plus (aux r)
      | Decompile.Trans (sset,qsub) -> Elem (sset, decompile qsub) in
    aux re
