(** Types (as states of tree automata) *)

open Tags

type t
val equal: t -> t -> bool
val hash: t -> int
val compare: t -> t -> int

val mk : unit -> t
  (** Create a fresh state which need to be later defined
      either by [set] or by [link]. *)

val set : t -> bool -> (SymbolSet.t * t * t) list -> unit
  (** Specify whether the state is final (the type contains
      the empty sequence), and its outgoing transitions. *)

val link: t -> t -> unit
  (** [link q1 q2] defines [q1] as an alias for [q2]. *)

val cons : bool -> (SymbolSet.t * t * t) list -> t
  (** [mk] and [set] at once. *)

val id: t -> int
val trans: t -> (SymbolSet.t * t * t) list
val final: t -> bool
val auto: t -> ((SymbolSet.t * t) * t) list * bool

val between : t -> t -> t list
  (** [between q1 q2] returns the states [q] accesible
      from [q1] and such that [q2] is accesible from [q]
      (with horizontal navigation). *)

val finals : t -> t list
  (** [finals q] returns the final states accessible from [q]. *)

val any: t
  (** The universal type. *)

val empty_seq: t
  (** The type of the empty sequence. *)

val alt : t -> t -> t
  (** Union of two types. If any of the two states is still undefined,
      it is considered as being empty. *)

val print: Format.formatter -> t -> unit
val print_def: Format.formatter -> t -> unit
val print_full: Format.formatter -> t -> unit

val is_empty: t -> bool
  (** Check whether a type is empty. *)

(*
val sample: t -> Syntax.t
  (** Extract a value from a type. Raises [Not_found] if the type
      is empty. *)
*)

val match_forest: t -> (Symbol.t * Syntax.t) list  -> bool
  (** Does the type matches the forest? *)
val match_val: t -> Syntax.t -> bool
  (** Does the type matches the value? *)


