open Tags

exception NotEmpty

let sample_elem sset sub next = 
  let e = Syntax.Elem (SymbolSet.finite [SymbolSet.sample sset], sub) in
  match next with
    | Syntax.Eps -> e
    | next -> Syntax.Seq (e,next)


type slot = { mutable status : status; 
              mutable notify : notify;
              mutable active : bool }
and status = Empty | NEmpty of Syntax.t | Maybe
and notify = Nothing | Do of slot * (Syntax.t -> unit) * notify

let slot_empty = { status = Empty; active = false; notify = Nothing }
let slot_not_empty = { status = NEmpty Syntax.Eps; 
		       active = false; notify = Nothing }

let rec notify v = function
  | Nothing -> ()
  | Do (n,f,rem) -> 
      if n.status == Maybe then (try f v with NotEmpty -> ());
      notify v rem

let set s v =
  s.status <- NEmpty v;
  notify v s.notify;
  s.notify <- Nothing; 
  raise NotEmpty

let guard a f n = match a with
  | { status = Empty } -> ()
  | { status = Maybe } as s -> n.active <- true; s.notify <- Do (n,f,s.notify)
  | { status = NEmpty v } -> f v

module StateSet = Slist.Make(State)

let memo = Hashtbl.create 8191
let marks = ref [] 

let ppf = Format.std_formatter

let rec slot t sl =
  if State.final t && not (List.exists State.final sl) then slot_not_empty
  else
    let key = (State.id t, List.map State.id sl) in
    try Hashtbl.find memo key
    with Not_found ->
(*      Format.fprintf ppf "Slot t=%a@." State.print t;
      List.iter (fun s ->
		   Format.fprintf ppf "  -> %a@." State.print s) sl; *)
		   
      let s = { status = Maybe; active = false; notify = Nothing } in
      Hashtbl.add memo key s;
      (try
	 List.iter 
	   (check_times s (List.flatten (List.map State.trans sl))) 
	   (State.trans t);
	 if s.active then marks := s :: !marks else s.status <- Empty
       with NotEmpty -> ());
      s
	   
  
and check_times s right (sset,q1,q2) =
  let rec aux sset v1 v2 accu1 accu2 right = match right with
    | [] -> set s (sample_elem sset v1 v2)
    | (sset',q1',q2')::right ->
	if SymbolSet.disjoint sset sset'
	  || StateSet.mem q1' accu1
	  || StateSet.mem q2' accu2  then aux sset v1 v2 accu1 accu2 right 
	else begin
	  (let sset = SymbolSet.diff sset sset' in
	   if not (SymbolSet.is_empty sset) then 
	     aux sset v1 v2 accu1 accu2 right);
	  if not (State.equal q1 q1') then
	    (let accu1 = StateSet.add q1' accu1 in
	     guard (slot q1 accu1) 
	       (fun v1 -> aux sset v1 v2 accu1 accu2 right) s);
	  if not (State.equal q2 q2') then
	    (let accu2 = StateSet.add q2' accu2 in
	     guard (slot q2 accu2) 
	       (fun v2 -> aux sset v1 v2 accu1 accu2 right) s)
	end
  in
  if not (SymbolSet.is_empty sset) then
    let accu1 = StateSet.empty and accu2 = StateSet.empty in
    guard (slot q1 accu1) 
      (fun v1 -> guard (slot q2 accu2) (
	 fun v2 -> aux sset v1 v2 accu1 accu2 right) s) s

let sample t tl =
  let s = slot t (StateSet.from_list tl) in
  List.iter 
    (fun s' -> 
       if s'.status == Maybe then s'.status <- Empty; s'.notify <- Nothing) 
    !marks;
  marks := [];
  match s.status with 
    | Empty -> None 
    | NEmpty v -> Some v 
    | Maybe -> assert false

let empty t tl =
  sample t tl == None

